# necessary imports
import sys
import os
import logging
import warnings
from pathlib import Path
from lark import Lark
import lark
from modules.process_tree import ProcessTree
from modules.fu_config import config
import argparse


THIS = Path(__file__).parent
# driver function
def driver():
    # set level to critical
    logging.getLogger().setLevel(logging.CRITICAL)
    warnings.filterwarnings('ignore')


    grammar = THIS/'grammar/grammar.lark';assert(grammar.is_file())

    with open(grammar,'r') as f: raw_grammar = f.read()
    # create the parser instance 
    parser = Lark(raw_grammar, parser="lalr", transformer=ProcessTree())

    # create argument parser
    args_parser = argparse.ArgumentParser(description='Interpreter for WDGAF language', epilog="And that's how you use the interepreter")
    args_parser.add_argument('src', metavar='src', help='source file path')
    args_parser.add_argument('-v', '--verbose', help="generates verbose output", action="store_true")
    args_parser.add_argument('-t','--tree',default=False,help='print parse tree',action ='store_true')

    # get the arguments
    args = args_parser.parse_args()

    # configure verbose parametere here!
    # this is used in all other modules!
    config['verbose'] = args.verbose

    # check if source file is provided
    if args.src: source = Path(args.src)
    else: print("Usage:\npython wdgaf.py <filename>"); sys.exit();

    # assert that file exist and it's *.fu file
    assert(source.is_file() and source.suffix == '.fu')
    # read the source file
    with open(source,'r') as f: code_raw = f.read()

    # execute user code
    try:
        if args.tree:
            printer = Lark(raw_grammar)
            print(printer.parse(code_raw).pretty())
        else:
            parser.parse(code_raw)

    # catch syntax errors
    except lark.exceptions.UnexpectedToken as e:
        token = e.token
        print("\033[91mSytnax Error: \033[00m", end='')
        print("unexpected token '{}' at line {}, column {}".format(token, token.line, token.column))
        print("process terminated")

    # these exceptions will generally be well-defined
    except Exception as ex:
        print("\033[91mError: \033[00m", end='')
        print(ex)
        print("process terminated")
        # set level to critical
        logging.getLogger().setLevel(logging.CRITICAL)
        warnings.filterwarnings('ignore')


        grammar = THIS/'grammar/grammar.lark';assert(grammar.is_file())

        with open(grammar,'r') as f: raw_grammar = f.read()
        # create the parser instance 
        parser = Lark(raw_grammar, parser="lalr", transformer=ProcessTree())

        # create argument parser
        args_parser = argparse.ArgumentParser(description='Interpreter for WDGAF language', epilog="And that's how you use the interepreter")
        args_parser.add_argument('src', metavar='src', help='source file path')
        args_parser.add_argument('-v', '--verbose', help="generates verbose output", action="store_true")
        args_parser.add_argument('-t','--tree',help='print parse tree')

        # get the arguments
        args = args_parser.parse_args()

        # configure verbose parametere here!
        # this is used in all other modules!
        config['verbose'] = args.verbose

        # check if source file is provided
        if args.src: source = Path(args.src)
        else: print("Usage:\npython wdgaf.py <filename>"); sys.exit();

        # assert that file exist and it's *.fu file
        assert(source.is_file() and source.suffix == '.fu')
        # read the source file
        with open(source,'r') as f: code_raw = f.read()

        # execute user code
        try:
            if args.tree:
                print(Lark(raw_grammar).parse(code_raw).pretty())
            else:
                parser.parse(code_raw)

        # catch syntax errors
        except lark.exceptions.UnexpectedToken as e:
                token = e.token
                print("\033[91mSytnax Error: \033[00m", end='')
                print("unexpected token '{}' at line {}, column {}".format(token, token.line, token.column))
                print("process terminated")

        # these exceptions will generally be well-defined
        except Exception as ex:
                print("\033[91mError: \033[00m", end='')
                print(ex)
                print("process terminated")

# invoke the driver function
if __name__ == '__main__':
    driver()



# necessary imports
import numpy as np

# parses the size
def parse_size(rows=None, cols=None):
	# parse the number of rows
	if rows:
		try: rows = int(rows)
		except: raise Exception("Number of rows must be a positive integer but found {}".format(rows))
		if rows < 1:
			raise Exception("Number of rows must be a positive integer but found {}".format(rows))
	else: rows = 1

	# parse the number of columns
	if cols:
		try: cols = int(cols)
		except: raise Exception("Number of cols must be a positive integer but found {}".format(cols))
		if cols < 1:
			raise Exception("Number of cols must be a positive integer but found {}".format(cols))
	else: cols = 1

	# return the parsed rows and cols
	return (rows, cols)



# parses the given list - (string or real)
def parse_list(type, points):
	# stores the list items
	items = []
	# iterate through every point
	for point in points:	
		# if this is to be converted to string
		# drop out additional quotations
		try:
			items.append(float(point))
		except:
			val = str(point)
			val = val[1:-1]
			items.append(val)
	# return the list
	return np.array(items)



# creates the range with given jump
def create_range_with_jump(start, end, jump, isInt):
	# take decision based on jump
	if not jump:
		# get the boundary
		j = 1 if isInt else 0.01
		# jump is not provided so we create range
		# based on bounds with jumps to be 1
		if start <= end: return np.arange(start, end + j, j)
		else: return np.arange(start, end - j, -j)
	else:
		# get the boundary
		j = 1 if isInt and jump.is_integer() else 0.01
		# jump is provided so we take care of it
		if jump == 0: raise Exception("step must be non-zero for range")
		elif jump < 0:
			# check if bounds are valid
			if start < end: raise Exception("start should be at least end with step {}".format(jump))
			# construct and return the items
			return np.arange(start, end - j, jump)
		else:
			# check if bounds are valid
			if start > end: raise Exception("start should be at most end with step {}".format(jump)) 
			# construct and return the items
			return np.arange(start, end + j, jump)

# creates range with given values
def create_range(start=None, end=None, jump=None):
	if start is None: raise Exception("start must be provided for range")
	if end is None: raise Exception("end must be provided for range")
	if jump == 0: raise Exception("step must be non-zero for range")
	# check if both the bounds are integer
	if start.is_integer() and end.is_integer():
		# create a range with integral bounds
		return create_range_with_jump(start, end, jump, True)
	else:
		# create a range with real bounds
		return create_range_with_jump(start, end, jump, False)





# necessary imports
import math
import numpy as np

# define the constants
CONSTANTS = {
	'PI' : np.pi,
	'EXP': np.e,
	'EULER_GAMMA': np.euler_gamma
}

# invokes the function on given item(s)
def invoke_unary_function(func, item):
	if func == 'random':
		if type(item) != float:
			raise Exception("Invalid parameter {} for random function".format(item))
		if not item.is_integer():
			raise Exception("Invalid parameter {} for random function".format(item))
		if item < 1:
			raise Exception("Invalid parameter {} for random function".format(item))
		return np.random.rand(int(item))

	# take decision based on func
	if 	 func == 'sin':			return np.sin(item)
	elif func == 'cos':			return np.cos(item)
	elif func == 'tan':			return np.tan(item)
	elif func == 'arcsin':		return np.arcsin(item)
	elif func == 'arccos':		return np.arccos(item)
	elif func == 'arctan':		return np.arctan(item)
	elif func == 'sinh':		return np.sinh(item)
	elif func == 'cosh':		return np.cosh(item)
	elif func == 'tanh':		return np.tanh(item)
	elif func == 'arcsinh':		return np.arcsinh(item)
	elif func == 'arccosh':		return np.arccosh(item)
	elif func == 'arctanh':		return np.arctanh(item)
	elif func == 'degrees':		return np.degrees(item)
	elif func == 'radians':		return np.radians(item)
	elif func == 'deg2rad':		return np.deg2rad(item)
	elif func == 'rad2deg':		return np.rad2deg(item)
	elif func == 'floor':		return np.floor(item)
	elif func == 'ceil':		return np.ceil(item)
	elif func == 'trunc':		return np.trunc(item)
	elif func == 'round':		return np.round(item)
	elif func == 'negative':	return np.negative(item)
	elif func == 'absolute':	return np.absolute(item)
	elif func == 'exp':			return np.exp(item)
	elif func == 'exp2':		return np.exp2(item)
	elif func == 'log':			return np.log(item)
	elif func == 'log2':		return np.log2(item)
	elif func == 'log10':		return np.log10(item)
	elif func == 'expm1':		return np.expm1(item)
	elif func == 'log1p':		return np.log1p(item)
	elif func == 'sqrt':		return np.sqrt(item)
	elif func == 'square':		return np.square(item)
	elif func == 'reciprocal':	return np.reciprocal(item)
	else	:					raise Exception("Unknown function {}".format(func))



# invokes the binary function on given item(s)
def invoke_binary_function(func, item1, item2):
	# take decision based on func
	if 	 func == 'add':			return np.add(item1, item2)
	elif func == 'subtract':	return np.subtract(item1, item2)
	elif func == 'multiply':	return np.multiply(item1, item2)
	elif func == 'divide': 		return np.divide(item1, item2)
	elif func == 'floor_divide': return np.floor_divide(item1, item2)
	elif func == 'power':		return np.power(item1, item2)
	elif func == 'mod': 		return np.mod(item1, item2)
	elif func == 'logaddexp':	return np.logaddexp(item1, item2)
	elif func == 'logaddexp2':	return np.logaddexp2(item1, item2)
	elif func == 'gcd': 		return np.gcd(item1, item2)
	elif func == 'lcm': 		return np.lcm(item1, item2)
	elif func == 'arctan2': 	return np.arctan2(item1, item2)
	elif func == 'hypot': 		return np.hypot(item1, item2)
	elif func == 'bitwise_and': return np.bitwise_and(item1, item2)
	elif func == 'bitwise_or':	return np.bitwise_or(item1, item2)
	elif func == 'bitwise_xor': return np.bitwise_xor(item1, item2)
	elif func == 'left_shift': 	return np.left_shift(item1, item2)
	elif func == 'right_shift':	return np.right_shift(item1, item2)
	elif func == 'maximum': 	return np.maximum(item1, item2)
	elif func == 'minimum':		return np.minimum(item1, item2)
	else :	raise Exception("Unknown function {}".format(func))



# parses the constant
def parse_constant(value):
	if value in CONSTANTS:	return CONSTANTS[value]
	else	:	raise Exception("Unknown constant {}".format(value))


# evaluates the reuslt for (item1 op item2)
def evaluate_op(item1, op, item2):
	if 		op == '+':	return np.add(item1, item2)
	elif 	op == "-":	return np.subtract(item1, item2)
	elif 	op == "*":	return np.multiply(item1, item2)
	elif 	op == "/":	return np.divide(item1, item2)
	elif 	op == "%":	return np.mod(item1, item2)
	elif 	op == "^":	return np.power(item1, item2)
	else 	:			raise Exception("Invalid operator {}".format(op))



# necessary imports
from matplotlib.colors import is_color_like
import matplotlib.pyplot as plt
import matplotlib
from modules.fu_config import size_attrs, plot_attrs

# parses the boolean string
def parse_bool(text):
	return (text.lower() == 'true')

# raises exception for illegal attributes
def raise_illegal_attribute_error(attr, value):
	raise Exception("Illegal attribute value {} for attribute {}".format(value, attr))

# creates the subplots and returns the items
def subplots(rows, cols, attr_list=None):
	if attr_list == None:
		# create subplots and return items
		return plt.subplots(rows, cols, 
					num="We Do Give A Figure")


	# make sure all the supplied attributes
	# are currently supported
	for key in attr_list.keys():
		if key not in size_attrs:
			raise Exception("Illegal attribute {} for size-statement".format(key))

	# parse the attributes one by one
	sharex = attr_list.get('sharex', 'none')
	if sharex not in ['true', 'false']:
		raise_illegal_attribute_error('sharex', sharex)

	sharey = attr_list.get('sharey', 'none')
	if sharey not in ['true', 'false']:
		raise_illegal_attribute_error('sharey', sharey)

	width = attr_list.get('width', 6.4)
	if type(width) != float or width <= 0:
		raise_illegal_attribute_error('width', width)
	height = attr_list.get('height', 4.8)
	if type(height) != float or width <= 0:
		raise_illegal_attribute_error('height', height)

	facecolor = attr_list.get('facecolor', 'white')
	if not is_color_like(facecolor):
		raise_illegal_attribute_error('facecolor', facecolor)
	edgecolor = attr_list.get('edgecolor', 'white')
	if not is_color_like(facecolor):
		raise_illegal_attribute_error('edgecolor', edgecolor)


	linewidth = attr_list.get('linewidth', 0.8)
	if type(linewidth) != float or linewidth < 0:
		raise_illegal_attribute_error('linewidth', linewidth)

	# create subplots and return items
	return plt.subplots(rows, cols, 
					num="We Do Give A Figure",
					sharex=parse_bool(sharex), 
					sharey=parse_bool(sharey),
					figsize=(width, height),
					facecolor=facecolor,
					edgecolor=edgecolor,
					linewidth=linewidth)


# returns the first empty-cell (if any otherwise last cell)
# index returned are 1-based
def first_empty_cell(grid, start_row=0, start_col=0):
	# get the number of rows and cols
	rows, cols = grid.shape
	# iterate through all cells
	for i in range(start_row, rows):
		for j in range(start_col, cols):
			# check if this satisify the criteria
			if grid[i][j] == 0:
				return (i + 1, j + 1)
	# return the last cell
	return (rows, cols)



# configures the additional arguments for plot
def configure_plot(ax, props):
	# parse the attributes one by one
	title = props.get('title', None)
	x_label = props.get('xlabel', None)
	y_label = props.get('ylabel', None)

	# parse the scales
	xscale = props.get('xscale', 'linear')
	if xscale not in ['linear', 'log', 'symlog', 'logit']:
		raise_illegal_attribute_error('xscale', xscale)
	yscale = props.get('yscale', 'linear')
	if yscale not in ['linear', 'log', 'symlog', 'logit']:
		raise_illegal_attribute_error('yscale', yscale)

	# parse the margins
	xmargin = props.get('xmargin', 0.1)
	if xmargin <= -0.5:
		raise_illegal_attribute_error('xmargin', xmargin)
	ymargin = props.get('ymargin', 0.1)
	if ymargin <= -0.5:
		raise_illegal_attribute_error('ymargin', ymargin)

	# scales
	xautoscale = props.get('xautoscale', 'true')
	if xautoscale not in ['true', 'false']:
		raise_illegal_attribute_error('xautoscale', xautoscale)
	yautoscale = props.get('yautoscale', 'true')
	if yautoscale not in ['true', 'false']:
		raise_illegal_attribute_error('yautoscale', yautoscale)

	# rotations
	xrotation = props.get('xrotation', 0.0)
	yrotation = props.get('yrotation', 0.0)

	# ticks on axis
	xmajorticks = props.get('xmajorticks', 'true')
	if xmajorticks not in ['true', 'false']:
		raise_illegal_attribute_error('xmajorticks', xmajorticks)
	ymajorticks = props.get('ymajorticks', 'true')
	if ymajorticks not in ['true', 'false']:
		raise_illegal_attribute_error('ymajorticks', ymajorticks)
	xminorticks = props.get('xminorticks', 'true')
	if xminorticks not in ['true', 'false']:
		raise_illegal_attribute_error('xminorticks', xminorticks)
	yminorticks = props.get('yminorticks', 'true')
	if yminorticks not in ['true', 'false']:
		raise_illegal_attribute_error('yminorticks', yminorticks)
	xticklabel = props.get('xticklabel', 'true')
	if xticklabel not in ['true', 'false']:
		raise_illegal_attribute_error('xticklabel', xticklabel)
	yticklabel = props.get('yticklabel', 'true')
	if yticklabel not in ['true', 'false']:
		raise_illegal_attribute_error('yticklabel', yticklabel)

	# gridlines
	gridlines = props.get('gridlines', 'false')
	if gridlines not in ['true', 'false']:
		raise_illegal_attribute_error('gridlines', gridlines)


	# fetch the facecolor
	facecolor = props.get('facecolor', 'white')
	if not is_color_like(facecolor): 
		raise_illegal_attribute_error('facecolor', facecolor)

	# set the properties
	ax.set_title(title)
	ax.set_xlabel(x_label)
	ax.set_ylabel(y_label)
	ax.set_xscale(xscale)
	ax.set_yscale(yscale)
	ax.set_xmargin(xmargin)
	ax.set_ymargin(ymargin)
	ax.set_autoscalex_on(parse_bool(xautoscale))
	ax.set_autoscaley_on(parse_bool(yautoscale))
	ax.tick_params(axis='x',rotation=xrotation)
	ax.tick_params(axis='x',which='major', bottom=parse_bool(xmajorticks))
	ax.tick_params(axis='x',which='minor', bottom=parse_bool(xminorticks))
	ax.tick_params(axis='x',labelbottom=parse_bool(xticklabel))
	ax.tick_params(axis='y',rotation=yrotation)
	ax.tick_params(axis='y',which='major', left=parse_bool(ymajorticks))
	ax.tick_params(axis='y',which='minor', left=parse_bool(yminorticks))
	ax.tick_params(axis='y',labelleft=parse_bool(yticklabel))
	ax.set_facecolor(facecolor)
	if parse_bool(gridlines): ax.grid()

# creates a plot with given values on given axes
def plot(xs, ys, zs, ax, props, category):
	# if no configuration is provided 
	# then use the default ones
	if props == None:
		if zs is not None: ax.contour(xs, ys, zs, [0])
		else:  ax.plot(xs, ys)

	# make sure all the supplied attributes
	# are currently supported
	for key in props.keys():
		if key not in plot_attrs:
			raise Exception("Illegal attribute {} for plot-statement".format(key))


	# take decision based on plot type
	if category == 'plot':
		# validate the additional parameters
		if 'level' in props: raise Exception("Illegal attribute level for plot-statement")
		# get the label for the plot
		label = props.get('label', None)
		# get the color
		color = props.get('color', None)
		if color != None and not is_color_like(color):
			raise_illegal_attribute_error('color', color)
		linestyle = props.get('linestyle', '-')
		if linestyle  not in ['-', '--', '-.', ':']:
			raise_illegal_attribute_error('linestyle', linestyle)
		linewidth = props.get('linewidth', 1.5)
		if linewidth <= 0.0:
			raise_illegal_attribute_error('linewidth', linewidth)
		fillstyle = props.get('fillstyle', 'full')
		if fillstyle  not in ['full', 'left', 'right', 'bottom', 'top', 'none']:
			raise_illegal_attribute_error('fillstyle', fillstyle)
		alpha = props.get('alpha', 1.0)
		if alpha < 0.0 or alpha > 1.0:
			raise raise_illegal_attribute_error('alpha', alpha)

		dashcapstyle = props.get('dashcapstyle', 'butt')
		if dashcapstyle not in ['butt', 'round', 'projecting']:
			raise_illegal_attribute_error('dashcapstyle', dashcapstyle)
		dashjoinstyle = props.get('dashjoinstyle', 'round')
		if dashjoinstyle not in ['miter', 'round', 'bevel']:
			raise_illegal_attribute_error('dashjoinstyle', dashjoinstyle)
		solidcapstyle = props.get('solidcapstyle', 'butt')
		if solidcapstyle not in ['butt', 'round', 'projecting']:
			raise_illegal_attribute_error('solidcapstyle', solidcapstyle)
		solidjoinstyle = props.get('solidjoinstyle', 'round')
		if solidjoinstyle not in ['miter', 'round', 'bevel']:
			raise_illegal_attribute_error('solidjoinstyle', solidjoinstyle)

		drawstyle = props.get('drawstyle', 'default')
		if drawstyle not in ['default', 'steps', 'steps-pre', 'steps-mid', 'steps-post']:
			raise raise_illegal_attribute_error('drawstyle', drawstyle)

		marker = props.get('marker', None)
		if marker and marker not in [".",",","o","v","^","<",">","1","2","3","4","8","s","p","P","*","h","H","+","x","X","D","d","|","_"]:
			raise_illegal_attribute_error('marker', marker)

		markeredgecolor = props.get('markeredgecolor', 'white')
		if not is_color_like(markeredgecolor):
			raise_illegal_attribute_error('markeredgecolor', markeredgecolor)
		markeredgewidth = props.get('markeredgewidth', 1.0)
		if markeredgewidth < 0.0:
			raise_illegal_attribute_error('markeredgewidth', markeredgewidth)
		markerfacecolor = props.get('markerfacecolor', 'white')
		if not is_color_like(markerfacecolor):
			raise_illegal_attribute_error('markerfacecolor', markerfacecolor)
		markersize = props.get('markersize', 6.0)
		if markersize < 0.0:
			raise_illegal_attribute_error('markersize', markersize)

		# create plot
		ax.plot(xs, ys, label=label, color=color, 
						linestyle=linestyle, linewidth=linewidth,
						fillstyle=fillstyle, alpha=alpha,
						dash_capstyle=dashcapstyle, dash_joinstyle=dashjoinstyle, 
						solid_capstyle=solidcapstyle, solid_joinstyle=solidjoinstyle, 
						drawstyle=drawstyle,
						marker=marker, markersize=markersize,
						markeredgecolor=markeredgecolor,
						markerfacecolor=markerfacecolor,
						markeredgewidth=markeredgewidth
				)
		# ax.plot(xs, ys)

	# contour plot
	elif category == 'contour':
		# validate the additional parameters
		if 'color' in props: raise Exception("Illegal attribute color for contour-statement")
		if 'label' in props: raise Exception("Illegal attribute label for contour-statement")
		# get the level	
		level = props.get('level', 'single')
		if level not in ['single', 'multiple']:
			raise_illegal_attribute_error('level', level)
		# plot the contour
		if level == "multiple": ax.contour(xs, ys, zs)
		else:	ax.contour(xs, ys, zs, [0])

	# configure the axis
	configure_plot(ax, props)




# creates the plot for the user
def create_plot(ax, grid, xs, ys, zs=None, attr_list=None, category=None):
	# fetch the number or rows and cols
	rows, cols = grid.shape
	# fetch the first non-empty cell
	def_row, def_col = first_empty_cell(grid)
	# fetch the index to plot
	row = attr_list.get('row', None)
	col = attr_list.get('col', None)
	# parse to integers
	if row != None:	row = int(row)
	if col != None: col = int(col)


	# if neither row nor col are provided we scan entire grid
	if row == None and col == None:
		# default ends to last cell
		row = rows; col = cols; found = False;
		# iterate through all cells
		for i in range(rows):
			for j in range(cols):
				if grid[i][j] == 0: col = j + 1; found=True; break;
			if found: row = i + 1; break;


	# if column but not row is provided we scan row
	elif row == None and col != None:		
		# check if column is valid
		if col < 1 or col > cols: 
			raise Exception("Column index out of range")
		# default ends to last row
		row = rows
		# iterate through all rows in this col
		for i in range(rows):
			if grid[i][col - 1] == 0: row = i + 1; break;


	# if row but not col is provided we scan col
	elif row != None and col == None:
		# check if row is valid		
		if row < 1 or row > rows: 
			raise Exception("Row index out of range")
		# default ends to last col
		col = cols
		# iterate through all cols in this row
		for j in range(cols):
			if grid[row - 1][j] == 0: col = j + 1; break;

	# check bounds
	if row < 1 or row > rows: raise Exception("Row index out of range")
	if col < 1 or col > cols: raise Exception("Column index out of range")

	# find appropriate indexing
	if rows == 1 and cols == 1:
		# axis cannot be subscript
		plot(xs, ys, zs, ax, attr_list, category)
	elif rows == 1 or cols == 1:
		# axis can be subscript using one index
		plot(xs, ys, zs, ax[row * col - 1], attr_list, category)
	else:
		# axis can be subscript using both index
		plot(xs, ys, zs, ax[row - 1][col - 1], attr_list, category)

	# occupy this location
	grid[row - 1][col - 1] = 1.0




# stores the verbose behavior
config = {
	'verbose': False
}

# defines the attributes for size-statement
size_attrs = [
	'sharex',
	'sharey',
	'width',
	'height',
	'facecolor',
	'edgecolor',
	'linewidth'
]

# defines the attributes for plot-statement
plot_attrs = [
	'label',
	'title',
	'color',
	'xlabel',
	'ylabel',
	'xscale',
	'yscale',
	'xmargin',
	'ymargin',
	'xautoscale',
	'yautoscale',
	'xrotation',
	'yrotation',
	'level',
	'xmajorticks',
	'ymajorticks',
	'xminorticks',
	'yminorticks',
	'xticklabel',
	'yticklabel',
	'gridlines',
	'facecolor',
	'marker',
	'linestyle',
	'linewidth',
	'alpha',
	'dashcapstyle',
	'dashjoinstyle',
	'solidcapstyle',
	'solidjoinstyle',
	'drawstyle',
	'fillstyle',
	'markeredgecolor',
	'markeredgewidth',
	'markerfacecolor',
	'markersize'
]



# necessary imports
from lark import Transformer, v_args
import matplotlib.pyplot as plt
import numpy as np

# custom modules
import modules.fu_math as fmath
import modules.fu_plot as fplot
import modules.fu_util as futil
from modules.fu_config import config
from modules.fu_keywords import KEYWORDS

# handles the evaluation using aliases
@v_args(inline=True)
class ProcessTree(Transformer):

	# creates subplots with given values
	def create_subplots(self, rows, cols, attr_list=None):
		self.rows = rows
		self.cols = cols
		self.grid = np.zeros((rows, cols))
		plt.close('all') # important before subplots
		self.fig, self.ax = fplot.subplots(rows, cols, attr_list)

	# constructor
	def __init__(self):
		# start with 1x1 grid
		self.create_subplots(1, 1)
		# start with an empty list of variables
		self.vars = {}



	########################### FIGURE STMT ###########
	# figure statement alias
	# stmt_list is irrelevant here
	# because all statements are processed separately
	def figure_stmt(self, fig_id=None, *stmt_list):
		# fig_id must be provided
		if not fig_id:	raise Exception("FigureId must be provided")
		# generate verbose message
		if config['verbose']: print("figId := {}".format(fig_id))
		# update the title
		self.fig.suptitle("[{}]".format(fig_id))
		# display the plots
		plt.tight_layout()
		plt.legend()
		plt.show()

		# reset-params
		self.create_subplots(1, 1)
		# clear all variables
		self.vars = {}



	########################### SIZE STMT ###########
	# size_row_stmt alias 
	def size_row_stmt(self, rows=None, attr_list=None):
		# parse the rows and cols
		rows, cols = futil.parse_size(rows, None)
		# create subplots
		self.create_subplots(rows, cols, attr_list)
		# generate verbose message
		if config['verbose']:
			print("rows := {}, cols := {}, attr_list := {}".format(rows, 1, attr_list))

	# size_row_col_stmt alias 
	def size_row_col_stmt(self, rows=None, 
							cols=None, attr_list=None):
		# parse the rows and cols
		rows, cols = futil.parse_size(rows, cols)
		# create subplots
		self.create_subplots(rows, cols, attr_list)
		# generate verbose message
		if config['verbose']:
			print("rows := {}, cols := {}, attr_list := {}".format(rows, cols, attr_list))



	########################### EXPRESSIONS ###########
	# expr_alias
	def expr(self, value):
		# grab and return the expression
		return value

	# expr_real alias
	def expr_real(self, number):
		# return the number
		return float(number)

	# expr_const alias
	def expr_const(self, const_item):
		# return the value of the constant
		return fmath.parse_constant(const_item)

	# expr_negation alias
	def expr_negation(self, value):
		# return the negation of expression
		return -1.0 * value
	# expr_positive alias
	def expr_positive(self, value):
		# return the value of expression
		return value

	# nested_expr alias
	def nested_expr(self, value):
		# return the value
		return value

	# expr_id alias
	def expr_id(self, var_id):
		# get the string representation from Token
		var_id = str(var_id)
		# check if this is a constant
		if var_id in fmath.CONSTANTS: return fmath.CONSTANTS[var_id]
		# check if value exist
		if var_id not in self.vars:	raise Exception("Unknown parameter {}".format(var_id))
		# fetch the list of values for this variable
		items = self.vars[var_id]
		# return the values
		return items

	# grab and apply appropriate operation
	def sum_expr(self, expr1, expr2):
		return np.add(expr1, expr2)
	def sub_expr(self, expr1, expr2):
		return np.subtract(expr1, expr2)
	def prd_expr(self, expr1, expr2):
		return np.multiply(expr1, expr2)
	def div_expr(self, expr1, expr2):
		return np.divide(expr1, expr2)
	def floor_div_expr(self, expr1, expr2):
		return np.floor_divide(expr1, expr2)
	def mod_expr(self, expr1, expr2):
		return np.mod(expr1, expr2)
	def exp_expr(self, expr1, expr2):
		return np.power(expr1, expr2)




	########################### RANGE STMT ###########
	# real_points_list alias 
	def real_points_list(self, *points):
		# parse and return the list
		return futil.parse_list(float, points)

	# str_points_list alias 
	def str_points_list(self, *points):
		# parse and return the list
		return futil.parse_list(str, points)

	# range_expr alias
	def range_expr(self, start=None, end=None, step=None):
		# create a range with give bounds
		return futil.create_range(start, end, step)

	# range_stmt alias
	def range_stmt(self, var_id, items=[]):
		# get the string representation from Token
		var_id = str(var_id)
		# add this variable
		self.vars[var_id] = items
		# generate verbose message
		if config['verbose']: print("{} := {}".format(var_id, items))
		# return the tuple containing var_id, items
		return (var_id, items)







	########################### FUNCTION STMT ###########
	# func_call_unary alias 
	def func_call_unary(self, func_name, item):
		# invoke the given function 
		return fmath.invoke_unary_function(func_name, item)

	# func_call_binary alias 
	def func_call_binary(self, func_name, item1, item2):
		# invoke the given function 
		return fmath.invoke_binary_function(func_name, item1, item2)

	# nested_func alias
	def nested_func(self, expression):
		# return the expression
		return expression

	# func_op_func alias
	def func_op_func(self, item1, op, item2=None):
		# evaluate the expression
		return fmath.evaluate_op(item1, op, item2)

	# fun_stmt alias
	def func_stmt(self, var_id, items):
		# get the string representation from Token
		var_id = str(var_id)
		# assign the values to the variable
		self.vars[var_id] = items
		# generate verbose message
		if config['verbose']: print("{} := {}".format(var_id, items))



	########################### PLOTTING ###########
	# print_stmt alias
	def print_stmt(self, *var_ids):
		# iterate through all variables
		for var_id in var_ids:
			# get the string representation from Token
			var_id = str(var_id)
			if var_id in self.vars:
				print("{} = {}".format(var_id, self.vars[var_id]))
			else:
				raise Exception("Unknown parameter {}".format(var_id))


	# plt_stmt alias	
	def plt_stmt(self, x_id, y_id, *attr_list):
		# get the string representation from Token
		x_id = str(x_id)
		y_id = str(y_id)
		# validate values
		if x_id not in self.vars:	raise Exception("Unknown parameter {}".format(x_id))
		if y_id not in self.vars:	raise Exception("Unknown parameter {}".format(y_id))
		# fetch values
		xs = self.vars[x_id]
		ys = self.vars[y_id]
		props = {}
		for item in attr_list: props = item
		fplot.create_plot(self.ax, self.grid, xs, ys, None, props, category='plot')


	########################### CONTOUR ###########
	def contour_grid(self, id_x, id_y, startx, endx, starty, endy):
		startx = float(startx)
		starty = float(starty)
		endx = float(endx)
		endy = float(endy)
		if startx > endx: raise Exception("startx must be at least endx for countour")
		if starty > endy: raise Exception("starty must be at least endy for countour")
		x = np.arange(startx, endx, 0.01)
		y = np.arange(starty, endy, 0.01)
		X, Y = np.meshgrid(x, y)
		self.vars[id_x] = X
		self.vars[id_y] = Y
		# generate verbose message
		if config['verbose']: print("{} := {}\n {} := {}".format(id_x, X, id_y, Y))

	def contour_stmt(self, x_id, y_id, items, *attr_list):
		# get the string representation from Token
		x_id = str(x_id)
		y_id = str(y_id)
		# validate values
		if x_id not in self.vars:	raise Exception("Unknown parameter {}".format(x_id))
		if y_id not in self.vars:	raise Exception("Unknown parameter {}".format(y_id))
		# fetch values
		xs = self.vars[x_id]
		ys = self.vars[y_id]
		props = {}
		for item in attr_list: props = item
		fplot.create_plot(self.ax, self.grid, xs, ys, items, props, category='contour')
		# generate verbose message
		if config['verbose']: print("eqn := {}".format(items))



	########################### OPERATOR ###########
	# grabs and returns the operator
	def operator(self, op):
		return op


	########################### ATTR LIST ###########
	# grabs and returns the attr_list
	def attr_list(self, *attrs):
		return dict(attrs)

	# grabs and returns the attribute
	def attribute(self, key, value):
		key, value = str(key), str(value)
		try: return (key, float(value))
		except: return (key, value[1:-1])



	########################### IDETIFIERS ###########
	# grabs the identifier and validate it
	def identifier(self, var_id):
		# check if this variable is a keyword
		if var_id in KEYWORDS:
			raise Exception("Illegal use of reserved word {}".format(var_id))
		# return the variable id
		return var_id




import sys
import os
import logging
import warnings
from pathlib import Path
from lark import Lark
import lark


sys.path.append(str(Path(__file__).parent))

from modules.process_tree import ProcessTree
from modules.fu_config import config

import wdgaf.gaf as gaf

import argparse


def main():
    # setup mechanism for any meta-requirements in the close future
    gaf.driver()

if __name__ == '__main__':
    main()



from pathlib import Path
from setuptools import setup,find_packages

ROOT = Path(__file__).parent

README = (ROOT/'wdgaf/README.md').read_text()

setup(
        name = "WDGAF",
        version = "1.0.6",
        description = "We Do Give a Figure",
        long_description = README,
        long_description_content_type = "text/markdown",
        url = "https://github.com/rajp152k/Compilers-II",
        author = "WDGAF",
        author_email = "rajp152k@gmail.com",
        license = "MIT",
        classifiers=[
                "License :: OSI Approved :: MIT License",
                "Programming Language :: Python :: 3",
                "Programming Language :: Python :: 3.8",
            ],
        packages = find_packages(where='.',exclude=('tests','docs')),
        include_package_data=True,
        install_requires=["matplotlib","numpy","lark-parser"],
        entry_points={
            "console_scripts":[
                "gaf=wdgaf.__main__:main"
                ]
            },
        )
